#include "lpac.h"

LPAC::LPAC(vector<int>& typed_list, vector<int>& impute_list,char *panel_data, char *rmb_data, vector<double *> &obs,int line_size, map<int,string> &rslist, int ptype){

   tlist  = typed_list;
   ilist = impute_list;  
   lsize =  line_size;
   rsmap = rslist;
   
   obsv = obs;
   
   //typed_map: give a SNP order, the map tells if the SNP is typed
   for(int i=0;i<tlist.size();i++)
     tmap[tlist[i]]=i;
   
   // default parameter values
   sigma = 1;
   delta = 0;
   
   pred_type=0;
   panel_type=ptype;
   
   thresh = 1e-4;       

   mutp = 1.5e-3;

      
   fprintf(stderr,"Reading panel data ...\n");
   parse_panel(panel_data);
   fprintf(stderr,"Reading recombination file ...\n");
   parse_recomb(rmb_data);
      
   fprintf(stderr, "Setting mean vectors ... \n");
   mean_o = gsl_matrix_calloc(tlist.size(),1);
   mean_u = gsl_matrix_calloc(ilist.size(),1);
 
   for(int i=0;i<ilist.size();i++)
     gsl_matrix_set(mean_u,i,0,compute_freq(ilist[i]));
   
   for(int i=0;i<tlist.size();i++)
     gsl_matrix_set(mean_o,i,0,compute_freq(tlist[i]));
     

}


// This function sets up covariance for predicting untyped SNP from start to end
void LPAC::setupCovar(int start, int end){
  //finding the leftmost typed SNP
  int tl = 0; // default as the first typed SNP
  for(int i=ilist[start];i>=0;i--){
     if(tmap.find(i)!=tmap.end()){
        tl = tmap[i];
        break;
      }
   }
   int tr = tlist.size()-1;	
    for(int i=ilist[end];i<(tlist.size()+ilist.size());i++)
     if(tmap.find(i)!=tmap.end()){
        tr = tmap[i];
        break;
      }
   //find relevent left side markers
   double coe = 1;
   int cur = ilist[start];
   for(int i=tl;i>=0;i--){
      coe *= compute_coef(tlist[i],cur);
      if(coe<thresh)
        break;
      else
        tl--;
        cur = tlist[i];
   }

   //find relevent right side markers
   coe = 1;
   cur = ilist[end];
   for(int i=tr;i<tlist.size();i++){
      coe *= compute_coef(cur,tlist[i]);
      if(coe<thresh)
        break;
      else
        tr++;
        cur = tlist[i];
   }

  // typed marker would be from tl to tr
  if(tl<0)
     tl=0;
   if(tr==tlist.size())
     tr = tlist.size()-1;
 
  int tnum = tr - tl +1;
  int inum = end-start+1; 


  fprintf(stderr,"Computing covariance matrix (oo)... %d x %d\n",tnum,tnum);

  //make Sigma_oo matrix
  cvar_oo = gsl_matrix_calloc(tnum,tnum);

  //fill the diagonal
  for(int i=tl;i<=tr;i++){
    double val = gsl_matrix_get(mean_o,i,0);
    double var = compute_D(tlist[i],tlist[i]);
    //double var = val*(1-val);
    gsl_matrix_set(cvar_oo,i-tl,i-tl,sigma*((1-mutp)*(1-mutp)*var + (mutp/2)*(1-mutp/2))+delta );
  }
  
  // off-diagonal
  for(int i=tl;i<=tr;i++){
    coe=1;
    cur = tlist[i];
    for(int j=i+1;j<=tr;j++){
      double val;
      if(coe<thresh)
        val = 0;
      else{
	val = (1-mutp)*(1-mutp)*compute_D(tlist[i],tlist[j]);
        coe *= compute_coef(cur,tlist[j]);
        cur = tlist[j];
        val *= coe;
      }
      gsl_matrix_set(cvar_oo,i-tl,j-tl,sigma*val);
      gsl_matrix_set(cvar_oo,j-tl,i-tl,sigma*val);
    }
  }
  fprintf(stderr,"Computing covariance matrix (uo)... %d x %d\n",inum,tnum);
  //make Sigma_uo matrix
  cvar_uo = gsl_matrix_calloc(inum,tnum);
 
  for(int i=start;i<=end;i++){
     double coe;
     for(int j=tl;j<=tr;j++){
       double val=0;
       coe = compute_coef(tlist[j],ilist[i]);
       if(coe>=thresh)
	 val = coe*compute_D(ilist[i],tlist[j])*(1-mutp)*(1-mutp);
       gsl_matrix_set(cvar_uo,i-start,j-tl,sigma*val);
     }
  }
 
  diff = gsl_matrix_calloc(tnum,lsize);
  
  for(int i=tl;i<=tr;i++)
    for(int j=0;j<lsize;j++){
      double val;
      if(obsv[i][j]<0)
          val = 0;
      else
	val = obsv[i][j] - (1.0 + pred_type)*( (1-mutp)*gsl_matrix_get(mean_o,i,0) + mutp/2 );
      
      gsl_matrix_set(diff, i-tl,j, val);
    }
  

}



void LPAC::predict(int start, int end){
 
  vector<int> oo_left_bound;
  vector<int> oo_right_bound;
  find_bounds(cvar_oo,oo_left_bound,oo_right_bound);


  gsl_matrix *cvar_oo_inv = ge_invert(cvar_oo,oo_left_bound, oo_right_bound);
  vector<int> uo_left_bound;
  vector<int> uo_right_bound;
  find_bounds(cvar_uo,uo_left_bound,uo_right_bound);

  fprintf(stderr,"Computing conditional mean ...\n");
 
  gsl_matrix *prod1 = sparse_matrix_mult1(cvar_uo, cvar_oo_inv,uo_left_bound,uo_right_bound);

  vector<int> prod1_lb;
  vector<int> prod1_rb;
  find_bounds(prod1,prod1_lb,prod1_rb);
  
  gsl_matrix *prod2 = sparse_matrix_mult1(prod1, diff,prod1_lb,prod1_rb);
  fprintf(stderr,"Computing conditional variance...\n");
  gsl_matrix *prod3 = sparse_matrix_mult2_diag(prod1,cvar_uo,prod1_lb,prod1_rb,uo_left_bound,uo_right_bound);
  for(int i=0;i<cvar_uo->size1;i++){
    //fprintf(outf,"%6d  %10s  ", start+i+1, rsmap[ilist[start+i]].c_str());
    fprintf(outf,"%10s  ", rsmap[ilist[start+i]].c_str());
    double mu = gsl_matrix_get(mean_u,i+start,0);
    double var = compute_D(ilist[start+i],ilist[start+i]);   
    //double var = mu*(1-mu);   
    double pred_var  = sigma*(var*(1-mutp)*(1-mutp) + (1-mutp/2)*mutp/2) + delta - gsl_matrix_get(prod3,i,0);
    for(int j=0;j<lsize;j++){
      double pred_mean = ((1-mutp)*mu + mutp/2)*(1.0 + pred_type) + gsl_matrix_get(prod2,i,j);
      if(pred_type==0){
	
        double ep = pred_mean;
        double ep2 = pred_mean*pred_mean + pred_var;
        double pg2 = ep2;
        double pg1 = 2*(ep-ep2);
        double pg0 = 1-pg1-pg2;
        //fprintf(outf,"%7.4f %7.4f     %7.4f %7.4f %7.4f ", pred_mean, sqrt(pred_var),pg0,pg1,pg2);
	fprintf(outf,"%7.4f   %7.4e   ", pred_mean, sqrt(pred_var));

	//printf("%7.4f %7.4e ", pred_mean, pred_var);
      }else
	fprintf(outf,"%7.4f  ",pred_mean);
    }
    fprintf(outf,"\n");
  }
  
  
  gsl_matrix_free(prod1);
  gsl_matrix_free(prod2);
  gsl_matrix_free(prod3);
  gsl_matrix_free(cvar_oo);
  gsl_matrix_free(cvar_oo_inv);
  gsl_matrix_free(cvar_uo);
  gsl_matrix_free(diff);

}


int LPAC::parse_panel(char *filename){
  
  FILE *fp = fopen(filename,"r");
  if(fp==0)
    return -1;

  vector<char*> content;
  vector<char>  line;

  while(1){
    char in = fgetc(fp);

    if(isalnum(in)){
      line.push_back(in);
      continue;
    }
    if(in=='\n'||in=='\r'||in==EOF){
      // convert line to single char array
      if(line.size()>0){
        char *pl = new char[line.size()+1];
        memset(pl,0,line.size()+1);
        for(int i=0;i<line.size();i++){
          pl[i]=line[i];
        }
        content.push_back(pl);
      }
      line.clear();
    }

    if(in==EOF)
      break;
  }

  fclose(fp);

  if(content.size()<1)
    return -2;

  //convert content to char**
  matrix = (char**)malloc(content.size()*sizeof(char*));
  memset(matrix,0,content.size());
  for(int i=0 ;i<content.size();i++){
    matrix[i]=content[i];
  }

  panel_size = content.size();
  content.clear();
  marker_num = strlen(matrix[0]);
  return marker_num;
}


int LPAC::parse_recomb (char * rmb_file){
  
  FILE *fp = fopen(rmb_file,"r");
  char str[81];
  while( fgets(str,80,fp) )
     rmb.push_back(atof(str));
  fclose(fp);
  return rmb.size();
}

double LPAC::compute_coef(int pos1, int pos2){

   if(pos2<pos1){
      int temp=pos1;
      pos1 = pos2;
      pos2 = temp;
   }
     
   double coe=1;
   for(int i=pos1+1;i<=pos2;i++)
     coe *= rmb[i];
   return coe;
}
   

double LPAC::compute_freq(int pos){
  int count = 0;
  int sum = 0;
  for(int i=0;i<panel_size;i++){
     int x = int(matrix[i][pos])-48;
     if(x>=0 && x<=2){
        count++;
        sum += x;
     }
  }
  double rst = double(sum)/double(count);
  if(panel_type==1)
    rst = rst/2.0;
  return rst;
}
     

double LPAC::compute_D(int pos1, int pos2){
  int sum = 0;
  int count = 0;
  int sumx = 0;
  int sumy = 0;
  for(int i=0;i<panel_size;i++){
     int x = int(matrix[i][pos1])-48;
     int y = int(matrix[i][pos2])-48;
     if(x<0||y<0||x>2||y>2)
       continue; 
     count++;
     sum += x*y;
     sumx += x;
     sumy += y;
  }
 
  double rst = double(sum)/double(count) - double(sumx*sumy)/double(count*count);
  if(panel_type==1)
    rst = rst/2.0;
  return rst;
}
  

// least square estimates for sigma^2 and delta^2
void LPAC::estimate_params(){
    
  // diagonal sums
  double sy_d, ss_d, ssy_d, ss2_d;
  sy_d = ss_d = ssy_d = ss2_d= 0;
  for(int i=0;i<tlist.size();i++){
    double val = gsl_matrix_get(mean_o,i,0);
    double var = compute_D(tlist[i],tlist[i]);
    //double var = val*(1-val); 
    double x = (1-mutp)*(1-mutp)*var + (mutp/2)*(1-mutp/2);
    double y = (obsv[i][0]-(1-mutp)*val-mutp/2)*(obsv[i][0]-(1-mutp)*val-mutp/2); 
    ss_d += x;
    sy_d += y;
    ssy_d += x*y;
    ss2_d += x*x;
  }
  
  //off-diagonal sums
  double ss2_o, ssy_o;  
  ss2_o = ssy_o =0;
    
  for( int i=0; i<tlist.size();i++){
    double val1 = obsv[i][0]-(1-mutp)*gsl_matrix_get(mean_o,i,0)-mutp/2;
    int cur = tlist[i];
    double coe = 1;
    for( int j=i+1; j<tlist.size();j++){
      coe *=  compute_coef(cur,tlist[j]);
      cur = tlist[j];    
      if(coe<thresh)
        break;
     
      double x = coe*(1-mutp)*(1-mutp)*compute_D(tlist[i],tlist[j]);
      double y = val1*(obsv[j][0]-(1-mutp)*gsl_matrix_get(mean_o,j,0)-mutp/2);  
      ss2_o += x*x;
      ssy_o += x*y;
    }
  }
  
  double ssy = ssy_d + ssy_o;
  double ss2 = ss2_d + ss2_o;

  sigma = (ssy*tlist.size()-sy_d*ss_d)/(ss2*tlist.size()-ss_d*ss_d);
  delta = (sy_d - ss_d*sigma)/tlist.size();
  if(delta<0){
    estimate_sigma(0);
  }
  
  fprintf(stderr, "Parameter estimation: sigma = %7.4e   delta = %7.4e\n",sigma,delta);
  
}



void LPAC::estimate_sigma(double input_delta){
  
  delta = input_delta;
  if(delta<0)
    delta = 0;
  // diagonal sums
  double  ss_d, ssy, ss2;
  ss_d = ssy = ss2= 0;
  for(int i=0;i<tlist.size();i++){
    double val = gsl_matrix_get(mean_o,i,0);
    double var= compute_D(tlist[i],tlist[i]);
    //double var = val*(1-val);
    double x = (1-mutp)*(1-mutp)*var + (mutp/2)*(1-mutp/2);
    double y = (obsv[i][0]-(1-mutp)*val-mutp/2)*(obsv[i][0]-(1-mutp)*val-mutp/2); 
    ss_d += x;
    ssy  += x*y;
    ss2  += x*x;
  }
  
  //off-diagonal sums
    
  for( int i=0; i<tlist.size();i++){
    double val1 = obsv[i][0]-(1-mutp)*gsl_matrix_get(mean_o,i,0)-mutp/2;
    int cur = tlist[i];
    double coe = 1;
    for( int j=i+1; j<tlist.size();j++){
      coe *=  compute_coef(cur,tlist[j]);
      cur = tlist[j];    
      if(coe<thresh)
        break;
     
      double x = coe*(1-mutp)*(1-mutp)*compute_D(tlist[i],tlist[j]);
      double y = val1*(obsv[j][0]-(1-mutp)*gsl_matrix_get(mean_o,j,0)-mutp/2);  
      ssy += x*y;
      ss2 += x*x;
    }
  }
  
  
  sigma = (ssy-delta*ss_d)/ss2;  

  fprintf(stderr, "Parameter estimation: sigma = %7.4e\n",sigma);
  
}


void LPAC::estimate_delta(double input_sigma){

  sigma = input_sigma;
  if(delta<0)
    delta = 0;
  // diagonal sums
  double  ss_d, sy_d;
  ss_d = sy_d = 0;
  for(int i=0;i<tlist.size();i++){
    double val = gsl_matrix_get(mean_o,i,0);
    double var = compute_D(tlist[i],tlist[i]);
    //double var = val*(1-val);
    double x = (1-mutp)*(1-mutp)*var + (mutp/2)*(1-mutp/2);
    double y = (obsv[i][0]-(1-mutp)*val-mutp/2)*(obsv[i][0]-(1-mutp)*val-mutp/2); 
    ss_d += x;
    sy_d += y;
  }
  
  delta = (sy_d - sigma*ss_d)/tlist.size();

  fprintf(stderr, "Parameter estimation: delta = %7.4e\n",delta);
}


void LPAC::set_outfile(char *outfile){
  if(outfile!=0){
    outf = fopen(outfile,"w");
    if(outf==0){
      fprintf(stderr,"Error: cannot open file %s for writing, use standard output instead\n\n",outfile);
      outf ==0;
    }
  }

  if(outf==0||outfile==0)
    outf = stdout;
}
